// This is an independent project of an individual developer. Dear PVS-Studio, please check it.
// PVS-Studio Static Code Analyzer for C, C++, C#, and Java: http://www.viva64.com
#include <string.h>
#include <errno.h>
#include "nrf.h"

#include "def.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"

#include "nrf_errno.h"
#include "bsd_os.h"
#include "bsd_platform.h"
#include "bsd_limits.h"

FILENUM(94441);

#define BSD_OS_TRACE_IRQ          EGU2_IRQn
#define BSD_OS_TRACE_IRQ_PRIORITY 6

#define BSD_APPLICATION_IRQ       EGU1_IRQn

static volatile TaskHandle_t sWaitingTaskHandleTbl[BSD_MAX_SOCKET_COUNT];
static volatile UInt32 sWaitingTaskCount = 0;


int32_t bsd_os_timedwait(uint32_t context, int32_t *timeout)
{
    // See https://github.com/nrfconnect/sdk-nrf/blob/master/lib/bsdlib/bsd_os.c
    UNUSED(context);

    TickType_t start = pdTICKS_TO_MS(xTaskGetTickCount());
    TickType_t sleepTime;

    // We can't do an infinite wait
    if(*timeout == 0)
    {
        return(NRF_ETIMEDOUT);
    }

    if((*timeout) < 0)
    {
        // wait forwever, or at least the maximum possible.
        sleepTime = portMAX_DELAY;
    }
    else
    {
        sleepTime = pdMS_TO_TICKS(*timeout);
    }


    /*
     * Add this tag to the list of sleeping tasks.
     */
    taskENTER_CRITICAL();
    size_t index = 0;
    // Loop in the table until you find the first emtpy (NULL) space.
    for(index = 0; (index < ARRAY_SIZE(sWaitingTaskHandleTbl)) && (NULL != sWaitingTaskHandleTbl[index]);
        index++)
    {
        continue;
    }
    if(index == ARRAY_SIZE(sWaitingTaskHandleTbl))
    {
        // No space, this should never happen.
        assert(false);
        NVIC_SystemReset();
    }
    sWaitingTaskHandleTbl[index] = xTaskGetCurrentTaskHandle();
    sWaitingTaskCount += 1;
    taskEXIT_CRITICAL();

    // Wait for bsd_os_application_irq_clear() to wakeup this thread.
    // return NRF_ETIMEOUT or remaining time depending on timer.
    ulTaskNotifyTake(pdFAIL, sleepTime);

    taskENTER_CRITICAL();
    sWaitingTaskHandleTbl[index] = NULL;
    sWaitingTaskCount -= 1;
    taskEXIT_CRITICAL();

    if(sleepTime == portMAX_DELAY)
    {
        return(0);
    }

    /* Calculate how much time is left until timeout. */
    UInt32 now = pdTICKS_TO_MS(xTaskGetTickCount());
    Int64 remaining = (Int64)(*timeout) - (now - start);

    *timeout = (remaining > 0) ? remaining : 0;
    if(*timeout == 0)
    {
        return(NRF_ETIMEDOUT);
    }

    return(0);
}



/// Function required by BSD library. We need to set the EGU1 interrupt.
void bsd_os_application_irq_set(void)
{
    NVIC_SetPendingIRQ(BSD_APPLICATION_IRQ);
}

/// Function required by BSD library. We need to clear the EGU1 interrupt.
void bsd_os_application_irq_clear(void)
{
    NVIC_ClearPendingIRQ(BSD_APPLICATION_IRQ);
}

/// Function required by BSD library. We need to set the EGU2 interrupt.
void bsd_os_trace_irq_set(void)
{
    NVIC_SetPendingIRQ(BSD_OS_TRACE_IRQ);
}

/// Function required by BSD library. We need to clear the EGU2 interrupt.
void bsd_os_trace_irq_clear(void)
{
    NVIC_ClearPendingIRQ(BSD_OS_TRACE_IRQ);
}

void read_task_create(void)
{
    // The read task is achieved using SW interrupt.
    NVIC_SetPriority(BSD_APPLICATION_IRQ, BSD_APPLICATION_IRQ_PRIORITY);
    NVIC_ClearPendingIRQ(BSD_APPLICATION_IRQ);
    NVIC_EnableIRQ(BSD_APPLICATION_IRQ);
}

void trace_task_create(void)
{
    NVIC_SetPriority(BSD_OS_TRACE_IRQ, BSD_OS_TRACE_IRQ_PRIORITY);
    NVIC_ClearPendingIRQ(BSD_OS_TRACE_IRQ);
    NVIC_EnableIRQ(BSD_OS_TRACE_IRQ);
}

/// Function required by BSD library.
void bsd_os_init(void)
{
    taskENTER_CRITICAL();
    memset((UInt8 *)sWaitingTaskHandleTbl, 0, sizeof(sWaitingTaskHandleTbl));
    sWaitingTaskCount = 0;
    taskEXIT_CRITICAL();

    read_task_create();
    trace_task_create();
}


/// Function required by BSD library
int32_t bsd_os_trace_put(const uint8_t *const data, uint32_t len)
{
    UNUSED(data);
    UNUSED(len);
    return(0);
}


/// Function required by BSD library. Stores an error code we can read later.
void bsd_os_errno_set(int err_code)
{
    // Translate nrf_errno.h errno to the OS specific value.
    errno = err_code;
}


void EGU1_IRQHandler(void)
{
    bsd_os_application_irq_handler();

    // Wakup sleeping/timeout threads.
    BaseType_t xHigherPriorityTaskWoken;

    UInt32 sleeperNotified = 0;
    size_t i = 0;
    while(sleeperNotified < sWaitingTaskCount)
    {
        assert(i < ARRAY_SIZE(sWaitingTaskHandleTbl));
        xHigherPriorityTaskWoken = pdFALSE;
        TaskHandle_t hdl = sWaitingTaskHandleTbl[i];
        if(hdl != null)
        {
            vTaskNotifyGiveFromISR(hdl, &xHigherPriorityTaskWoken);
            sleeperNotified += 1;
        }
        i += 1;
    }

    portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}

void EGU2_IRQHandler(void)
{
    bsd_os_trace_irq_handler();
}
